@@ -2,8 +2,6 @@
 # Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
 # MIT License (see LICENSE or https://opensource.org/licenses/MIT)
 
-from __future__ import division
-
 import binascii
 import hashlib
 import hmac
@@ -49,11 +47,11 @@ from smbprotocol import (
 )
 
 from smbprotocol._text import (
-    to_native,
     to_text,
 )
 
 from smbprotocol.exceptions import (
+    SMBConnectionClosed,
     SMB2SymbolicLinkErrorResponse,
     SMBException,
     SMBResponseException,
@@ -85,6 +83,7 @@ from smbprotocol.structure import (
 )
 
 from smbprotocol.transport import (
+    _TimeoutError,
     Tcp,
 )
 
@@ -610,14 +609,6 @@ class SMB2TransformHeader(Structure):
         super(SMB2TransformHeader, self).__init__()
 
 
-def _worker_running(func):
-    """ Ensures the message worker thread is still running and hasn't failed for any reason. """
-    def wrapped(self, *args, **kwargs):
-        self._check_worker_running()
-        return func(self, *args, **kwargs)
-    return wrapped
-
-
 class Connection(object):
 
     def __init__(self, guid, server_name, port=445, require_signing=True):
@@ -724,9 +715,9 @@ class Connection(object):
             negotiation process to complete
         """
         log.info("Setting up transport connection")
-        message_queue = Queue()
-        self.transport = Tcp(self.server_name, self.port, message_queue, timeout)
-        t_worker = threading.Thread(target=self._process_message_thread, args=(message_queue,),
+        self.transport = Tcp(self.server_name, self.port, timeout)
+        self.transport.connect()
+        t_worker = threading.Thread(target=self._process_message_thread,
                                     name="msg_worker-%s:%s" % (self.server_name, self.port))
         t_worker.daemon = True
         t_worker.start()
@@ -747,7 +738,10 @@ class Connection(object):
                     SecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED):
             self.require_signing = True
         log.info("Connection require signing: %s" % self.require_signing)
+
         capabilities = smb_response['capabilities']
+        self.server_capabilities = capabilities
+        self.server_security_mode = smb_response['security_mode'].get_value()
 
         # SMB 2.1
         if self.dialect >= Dialects.SMB_2_1_0:
@@ -768,9 +762,6 @@ class Connection(object):
             self.supports_encryption = capabilities.has_flag(
                 Capabilities.SMB2_GLOBAL_CAP_ENCRYPTION) \
                 and self.dialect < Dialects.SMB_3_1_1
-            self.server_capabilities = capabilities
-            self.server_security_mode = \
-                smb_response['security_mode'].get_value()
 
             # TODO: Check/add server to server_list in Client Page 203
 
@@ -796,14 +787,16 @@ class Connection(object):
         :param close: Will close all sessions in the connection as well as the
             tree connections of each session.
         """
-        if close:
+        # We cannot close the session or tree if the socket has been closed.
+        if close and self.transport.connected:
             for session in list(self.session_table.values()):
                 session.disconnect(True)
 
         log.info("Disconnecting transport connection")
         self.transport.close()
 
-    def send(self, message, sid=None, tid=None, credit_request=None, message_id=None, async_id=None):
+    def send(self, message, sid=None, tid=None, credit_request=None, message_id=None, async_id=None,
+             force_signature=False):
         """
         Will send a message to the server that is passed in. The final unencrypted header is returned to the function
         that called this.
@@ -814,10 +807,11 @@ class Connection(object):
         :param credit_request: Specifies extra credits to be requested with the SMB header.
         :param message_id: The message_id for the header, only useful for a cancel request.
         :param async_id: The async_id for the header, only useful for a cancel request.
+        :param force_signature: Force signing the SMB request even if not requested by the client/server.
         :return: Request of the message that was sent.
         """
         return self._send([message], session_id=sid, tree_id=tid, message_id=message_id, credit_request=credit_request,
-                          async_id=async_id)[0]
+                          async_id=async_id, force_signature=force_signature)[0]
 
     def send_compound(self, messages, sid, tid, related=False):
         """
@@ -834,7 +828,6 @@ class Connection(object):
         """
         return self._send(messages, session_id=sid, tree_id=tid, related=related)
 
-    @_worker_running
     def receive(self, request, wait=True, timeout=None, resolve_symlinks=True):
         """
         Polls the message buffer of the TCP connection and waits until a valid
@@ -847,6 +840,9 @@ class Connection(object):
         :param resolve_symlinks: Set to automatically resolve symlinks in the path when opening a file or directory.
         :return: SMB2HeaderResponse of the received message
         """
+        # Make sure the receiver is still active, if not this raises an exception.
+        self._check_worker_running()
+
         start_time = time.time()
         while True:
             iter_timeout = int(max(timeout - (time.time() - start_time), 1)) if timeout is not None else None
@@ -925,15 +921,16 @@ class Connection(object):
                         raise error
 
                     return new_response
-                elif status not in [NtStatus.STATUS_SUCCESS, NtStatus.STATUS_PENDING]:
-                    raise SMBResponseException(response)
                 else:
-                    break
+                    # now we have a retrieval request for the response, we can delete
+                    # the request from the outstanding requests
+                    message_id = request.message['message_id'].get_value()
+                    self.outstanding_requests.pop(message_id, None)
 
-        # now we have a retrieval request for the response, we can delete
-        # the request from the outstanding requests
-        message_id = request.message['message_id'].get_value()
-        del self.outstanding_requests[message_id]
+                    if status not in [NtStatus.STATUS_SUCCESS, NtStatus.STATUS_PENDING]:
+                        raise SMBResponseException(response)
+
+                    break
 
         return response
 
@@ -997,7 +994,7 @@ class Connection(object):
         actual = header['signature'].get_value()
         if actual != expected:
             raise SMBException("Server message signature could not be verified: %s != %s"
-                               % (to_native(binascii.hexlify(actual)), to_native(binascii.hexlify(expected))))
+                               % (binascii.hexlify(actual).decode(), binascii.hexlify(expected).decode()))
 
     def _check_worker_running(self):
         """ Checks that the message worker thread is still running and raises it's exception if it has failed. """
@@ -1005,9 +1002,11 @@ class Connection(object):
             self.disconnect(False)
             raise self._t_exc
 
-    @_worker_running
+        elif not self.transport.connected:
+            raise SMBConnectionClosed('SMB socket was closed, cannot send or receive any more data')
+
     def _send(self, messages, session_id=None, tree_id=None, message_id=None, credit_request=None, related=False,
-              async_id=None):
+              async_id=None, force_signature=False):
         send_data = b""
         requests = []
         session = self.session_table.get(session_id, None)
@@ -1067,7 +1066,7 @@ class Connection(object):
                 header['tree_id'] = b"\xff" * 4
                 header['flags'].set_flag(Smb2Flags.SMB2_FLAGS_RELATED_OPERATIONS)
 
-            if session and session.signing_required and session.signing_key:
+            if force_signature or (session and session.signing_required and session.signing_key):
                 header['flags'].set_flag(Smb2Flags.SMB2_FLAGS_SIGNED)
                 b_header = header.pack() + padding
                 signature = self._generate_signature(b_header, session.signing_key)
@@ -1102,32 +1101,33 @@ class Connection(object):
         if session and session.encrypt_data or tree and tree.encrypt_data:
             send_data = self._encrypt(send_data, session)
 
+        self._check_worker_running()
         self.transport.send(send_data)
         return requests
 
-    def _process_message_thread(self, msg_queue):
-        while True:
-            # Wait for a max of 10 minutes before sending an echo that tells the SMB server the client is still
-            # available. This stops the server from closing the connection and the associated sessions on a long lived
-            # connection. A brief test shows Windows kills a connection at ~16 minutes so 10 minutes is a safe choice.
-            # https://github.com/jborean93/smbprotocol/issues/31
-            try:
-                b_msg = msg_queue.get(timeout=600)
-            except Empty:
-                log.debug("Sending SMB2 Echo to keep connection alive")
-                for sid in self.session_table.keys():
-                    req = self.send(SMB2Echo(), sid=sid)
-                    # Set this reserved field to 1 as we use that internally to check whether the outstanding requests
-                    # queue should be cleared in this thread or not.
-                    req.message['reserved'] = 1
+    def _process_message_thread(self):
+        try:
+            while True:
+                # Wait for a max of 10 minutes before sending an echo that tells the SMB server the client is still
+                # available. This stops the server from closing the connection and the associated sessions on a long
+                # lived connection. A brief test shows Windows kills a connection at ~16 minutes so 10 minutes is a
+                # safe choice.
+                # https://github.com/jborean93/smbprotocol/issues/31
+                try:
+                    b_msg = self.transport.recv(600)
+                except _TimeoutError:
+                    log.debug("Sending SMB2 Echo to keep connection alive")
+                    for sid in self.session_table.keys():
+                        req = self.send(SMB2Echo(), sid=sid)
+                        # Set this reserved field to 1 as we use that internally to check whether the outstanding
+                        # requests queue should be cleared in this thread or not.
+                        req.message['reserved'] = 1
+                    continue
 
-                continue
+                # If recv didn't return any data then the socket is considered to be closed.
+                if not b_msg:
+                    return
 
-            # The socket will put None in the queue if it is closed, signalling the end of the connection.
-            if b_msg is None:
-                return
-
-            try:
                 is_encrypted = b_msg[:4] == b"\xfdSMB"
                 if is_encrypted:
                     msg = SMB2TransformHeader()
@@ -1187,19 +1187,19 @@ class Connection(object):
                         # When we send a ping in this thread we want to make sure it doesn't linger in the outstanding
                         # request queue.
                         if request.message['reserved'].get_value() == 1:
-                            del self.outstanding_requests[message_id]
-            except Exception as exc:
-                # The exception is raised in _check_worker_running by the main thread when send/receive is called next.
-                self._t_exc = exc
+                            self.outstanding_requests.pop(message_id, None)
+        except Exception as exc:
+            # The exception is raised in _check_worker_running by the main thread when send/receive is called next.
+            self._t_exc = exc
 
-                # Make sure we fire all the request events to ensure the main thread isn't waiting on a receive.
-                for request in self.outstanding_requests.values():
-                    request.response_event.set()
+            # While a caller of send/receive could theoretically catch this exception, we consider any failures
+            # here a fatal errors and the connection should be closed so we exit the worker thread.
+            self.disconnect(False)
 
-                # While a caller of send/receive could theoretically catch this exception, we consider any failures
-                # here a fatal errors and the connection should be closed so we exit the worker thread.
-                self.disconnect(False)
-                return
+        finally:
+            # Make sure we fire all the request events to ensure the main thread isn't waiting on a receive.
+            for request in self.outstanding_requests.values():
+                request.response_event.set()
 
     def _generate_signature(self, b_header, signing_key):
         b_header = b_header[:48] + (b"\x00" * 16) + b_header[64:]
@@ -1301,11 +1301,19 @@ class Connection(object):
                       % self.client_guid)
             neg_req['client_guid'] = self.client_guid
 
+        else:
+            # Must be None, this value is used to verify the negotiation info.
+            self.client_guid = None
+
         if highest_dialect >= Dialects.SMB_3_0_0:
             log.debug("Adding client capabilities %d to negotiate request"
                       % self.client_capabilities)
             neg_req['capabilities'] = self.client_capabilities
 
+        else:
+            # Must be 0, this value is used to verify the negotiation info.
+            self.client_capabilities = 0
+
         if highest_dialect >= Dialects.SMB_3_1_1:
             int_cap = SMB2NegotiateContextRequest()
             int_cap['context_type'] = \
@@ -1366,33 +1374,30 @@ class Connection(object):
         :param message: The message being sent
         :return: The credit charge to set on the header
         """
-        credit_size = MAX_PAYLOAD_SIZE
-
         if (not self.supports_multi_credit) or (message.COMMAND == Commands.SMB2_CANCEL):
-            credit_charge = 0
+            return 0
+
         elif message.COMMAND == Commands.SMB2_READ:
-            max_size = message['length'].get_value() + \
-                       message['read_channel_info_length'].get_value() - 1
-            credit_charge = math.ceil(max_size / credit_size)
+            payload_size = message['length'].get_value() + message['read_channel_info_length'].get_value()
+
         elif message.COMMAND == Commands.SMB2_WRITE:
-            max_size = message['length'].get_value() + \
-                       message['write_channel_info_length'].get_value() - 1
-            credit_charge = math.ceil(max_size / credit_size)
+            payload_size = message['length'].get_value() + message['write_channel_info_length'].get_value()
+
         elif message.COMMAND == Commands.SMB2_IOCTL:
             max_in_size = len(message['buffer'])
             max_out_size = message['max_output_response'].get_value()
-            max_size = max(max_in_size, max_out_size) - 1
-            credit_charge = math.ceil(max_size / credit_size)
+            payload_size = max(max_in_size, max_out_size)
+
         elif message.COMMAND == Commands.SMB2_QUERY_DIRECTORY:
             max_in_size = len(message['buffer'])
             max_out_size = message['output_buffer_length'].get_value()
-            max_size = max(max_in_size, max_out_size) - 1
-            credit_charge = math.ceil(max_size / credit_size)
-        else:
-            credit_charge = 1
+            payload_size = max(max_in_size, max_out_size)
 
-        # python 2 returns a float where we need an integer
-        return int(credit_charge)
+        else:
+            payload_size = 1
+
+        credit_charge = (max(0, payload_size - 1) // MAX_PAYLOAD_SIZE) + 1
+        return credit_charge
 
 
 class Request(object):
